#include <strsafe.h>
#include <windows.h>
#include <wchar.h>
#include <AtlBase.h>
#include <AtlConv.h>

void printToConsole(LPCWSTR displayedBuffer) {
  HANDLE stdOut = GetStdHandle(STD_OUTPUT_HANDLE);
  if (stdOut != NULL && stdOut != INVALID_HANDLE_VALUE)
  {
    DWORD out;
    WriteConsoleW(
      stdOut, displayedBuffer, lstrlenW(displayedBuffer), &out,
      NULL);
  }
}

void printError(LPWSTR function) {
  // Retrieve the system error message for the last-error code
  DWORD errorID = GetLastError();

  if (errorID == 0) {
    return; //No error message has been recorded
  }

  LPWSTR messageBuffer = nullptr;

    //Ask Win32 to give us the string version of that message ID.
    //The parameters we pass in, tell Win32 to create the buffer that holds the message for us (because we don't yet know how long the message string will be).
    size_t size = FormatMessage(
      FORMAT_MESSAGE_ALLOCATE_BUFFER
      | FORMAT_MESSAGE_FROM_SYSTEM
      | FORMAT_MESSAGE_IGNORE_INSERTS,
      NULL,
      errorID,
      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
      (LPWSTR)&messageBuffer,
      0, NULL);

    LPVOID displayedBuffer;
    displayedBuffer = (LPVOID)LocalAlloc(
      LMEM_ZEROINIT,
      (lstrlen((LPWSTR)messageBuffer) + lstrlen((LPCTSTR)function) + 40) * sizeof(WCHAR));
    StringCchPrintf((LPWSTR)displayedBuffer,
      LocalSize(displayedBuffer) / sizeof(WCHAR),
      TEXT("%s failed with error %d: %s"),
      function, errorID, messageBuffer);
    printToConsole((LPWSTR)displayedBuffer);
    // Free the Win32's string's buffer.
    LocalFree(messageBuffer);
    LocalFree(displayedBuffer);
}

int wmain(int argc, wchar_t const* argv[]) {
  if (argc != 2) {
    return 1;
  }

  wchar_t const* path = argv[1];
  HANDLE h = CreateFileW(
    path,
    0,
    (FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE),
    NULL,
    OPEN_EXISTING,
    FILE_FLAG_BACKUP_SEMANTICS,
    NULL);
  if (h == INVALID_HANDLE_VALUE) {
    printError(CA2T("CreateFile"));
    return 1;
  }

  wchar_t buff[MAX_PATH];
  WCHAR* lpBuf = buff;
  DWORD len = GetFinalPathNameByHandleW(h, buff, MAX_PATH, 0);
  if (len >= MAX_PATH) {
    lpBuf = (WCHAR*)malloc(len * sizeof(WCHAR));
    len = GetFinalPathNameByHandleW(h, lpBuf, len, 0);
  }
  CloseHandle(h);

  if (len == 0) {
    printError(CA2T("GetFinalPathNameByHandle"));
    return 1;
  }

  printToConsole(lpBuf);

  if (buff != lpBuf) {
    free(lpBuf);
  }
  return 0;
}
